package tech.riemann.ims.controller.platform.material;

import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.Parameter;
import io.swagger.v3.oas.annotations.tags.Tag;
import lombok.RequiredArgsConstructor;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.validation.annotation.Validated;
import org.springframework.web.bind.annotation.*;
import tech.riemann.ims.dto.request.ApplyInfo;
import tech.riemann.ims.dto.request.AuditApplyInfo;
import tech.riemann.ims.dto.response.ApplyDTO;
import tech.riemann.ims.entity.material.*;
import tech.riemann.ims.enums.*;
import tech.riemann.ims.service.material.*;

import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * @author mayong
 * @since 2024/11/28 15:29
 */
@RestController
@Tag(name = "Apply", description = "入库/出库 申请单管理")
@RequiredArgsConstructor
public class ApplyFormController {

    private final IApplyFormService applyFormService;
    private final IApplyDetailService applyDetailService;
    private final IMaterialStockDetailService materialStockDetailService;
    private final IMaterialService materialService;
    private final IStocktakingScanDetailService stocktakingScanDetailService;
    private final IStocktakingScanExceptionalDataService stocktakingScanExceptionalDataService;


    @GetMapping("applies")
    @Operation(summary = "分页查询申请列表")
    public IPage<ApplyDTO> searchPage(
            @Parameter(description = "页码") @RequestParam(name = "page", required = false, defaultValue = "1") int page,
            @Parameter(description = "页面大小") @RequestParam(name = "size", required = false, defaultValue = "10") int size,
            @Parameter(description = "申请类型(1入库 3出库 4盘点)") @RequestParam(name = "applyType") Integer applyType,
            @Parameter(description = "物料类型") @RequestParam(name = "type", required = false) String type,
            @Parameter(description = "物料编码") @RequestParam(name = "code", required = false) String code,
            @Parameter(description = "物料名称") @RequestParam(name = "name", required = false) String name) {
        return applyFormService.search(page, size, applyType, type, code, name);
    }

    @GetMapping("audit-applies")
    @Operation(summary = "分页查询盘点列表")
    public IPage<ApplyForm> searchAuditPage(
            @Parameter(description = "页码") @RequestParam(name = "page", required = false, defaultValue = "1") int page,
            @Parameter(description = "页面大小") @RequestParam(name = "size", required = false, defaultValue = "10") int size,
            @Parameter(description = "盘点类型") @RequestParam(name = "auditType", required = false) AuditTypeEnum auditType,
            @Parameter(description = "盘点人") @RequestParam(name = "taker", required = false) String taker,
            @Parameter(description = "创建日期") @RequestParam(name = "createDate", required = false) LocalDateTime createDate,
            @Parameter(description = "审核状态") @RequestParam(name = "reviewResults", required = false) List<ReviewResultEnum> reviewResults) {
        return applyFormService.page(Page.of(page, size), Wrappers.<ApplyForm>lambdaQuery()
                .eq(auditType != null, ApplyForm::getAuditType, auditType)
                        .eq(ApplyForm::getType, ApplyTypeEnum.AUDIT)
                .eq(taker != null, ApplyForm::getTaker, taker)
                .eq(createDate != null, ApplyForm::getCreatedTime, createDate)
                .in(reviewResults != null && !reviewResults.isEmpty(), ApplyForm::getReviewResult, reviewResults)
                .orderByDesc(ApplyForm::getCreatedTime)
        );
    }

    @GetMapping("apply/{applyId}")
    @Operation(summary = "查询申请单详情")
    public ApplyInfo detail(@Parameter(description = "申请单ID") @PathVariable(name = "applyId") Integer applyId) {
        List<ApplyDetail> list = applyDetailService.list(Wrappers.<ApplyDetail>lambdaQuery().eq(ApplyDetail::getApplyId, applyId));
        return ApplyInfo.builder()
                .applyForm(applyFormService.getById(applyId))
                .applyDetails(list)
                .build();
    }

    @PutMapping("apply")
    @Operation(summary = "提交出入库申请单")
    @Transactional(rollbackFor = Exception.class)
    public void saveApply(@Validated @Parameter(description = "申请单") @RequestBody ApplyInfo applyInfo) {
        // 如果是采购入库 就新增数据，否则只改状态
        ApplyTypeEnum applyType = applyInfo.getApplyForm().getType();

        // 校验物料数量
        Map<Long, Integer> confirmMap = applyInfo.getApplyDetails().stream()
                .filter(item -> item.getAssignRule() != AssignRuleEnum.LOW_VALUE)
                .collect(Collectors.toMap(ApplyDetail::getMaterialId, ApplyDetail::getConfirmQuantity));
        Map<Long, Long> dataMap = applyInfo.getDetailList().stream()
                .collect(Collectors.groupingBy(MaterialStockDetail::getMaterialId, Collectors.counting()));
        confirmMap.forEach((k, v) -> {
            Long l = dataMap.get(k);
            if (l == null || l.intValue() != v) {
                throw new IllegalArgumentException("[" + k + "]物料数量不匹配");
            }
        });
        // 新增申请单
        applyFormService.save(applyInfo.getApplyForm());
        for (ApplyDetail applyDetail : applyInfo.getApplyDetails()) {
            applyDetail.setApplyId(applyInfo.getApplyForm().getId());
        }
        applyDetailService.saveBatch(applyInfo.getApplyDetails());

        if (applyType == ApplyTypeEnum.PURCHASE_RECEIPT) {
            // 新增物料
            if(!applyInfo.getDetailList().isEmpty()){
                applyInfo.getDetailList().forEach(detail -> {
                    detail.setApplyId(applyInfo.getApplyForm().getId());
                    detail.setStatus(StockStatusEnum.IN);
                });
                materialStockDetailService.saveBatch(applyInfo.getDetailList());
            }
            //记录物料库存数据
            applyInfo.getApplyDetails().forEach(detail ->
                    materialService.update(Wrappers.<Material>lambdaUpdate()
                            .eq(Material::getId, detail.getMaterialId())
                            .setSql(detail.getAssignRule() == AssignRuleEnum.HIGH_VALUE,
                                    "m_stock = m_stock + " + detail.getConfirmQuantity())
                            .setSql(detail.getAssignRule() == AssignRuleEnum.LOW_VALUE,
                                    "m_stock = m_stock + " + detail.getQuantity())
                    ));
        } else if (applyType == ApplyTypeEnum.LOAN_OUT) {
            //修改物料状态
            if(!applyInfo.getDetailList().isEmpty()){
                applyInfo.getDetailList().forEach(detail -> materialStockDetailService.update(Wrappers.<MaterialStockDetail>lambdaUpdate()
                        .set(MaterialStockDetail::getStatus, StockStatusEnum.OUT)
                        .eq(MaterialStockDetail::getBarcode, detail.getBarcode())));
            }
            //修改物料库存数据
            applyInfo.getApplyDetails().forEach(detail -> materialService.update(Wrappers.<Material>lambdaUpdate()
                    .eq(Material::getId, detail.getMaterialId())
                    .setSql(detail.getAssignRule() == AssignRuleEnum.HIGH_VALUE,
                            "m_stock = m_stock - " + detail.getConfirmQuantity())
                    .setSql(detail.getAssignRule() == AssignRuleEnum.LOW_VALUE,
                            "m_stock = m_stock - " + detail.getQuantity())
            ));
        } else if (applyType == ApplyTypeEnum.RETURN_RECEIPT) {
            //修改物料状态
            if(!applyInfo.getDetailList().isEmpty()){
                applyInfo.getDetailList().forEach(detail -> materialStockDetailService.update(Wrappers.<MaterialStockDetail>lambdaUpdate()
                        .set(MaterialStockDetail::getStatus, StockStatusEnum.IN)
                        .eq(MaterialStockDetail::getBarcode, detail.getBarcode())));
            }
            //修改物料库存数据
            applyInfo.getApplyDetails().forEach(detail -> materialService.update(Wrappers.<Material>lambdaUpdate()
                    .eq(Material::getId, detail.getMaterialId())
                    .setSql(detail.getAssignRule() == AssignRuleEnum.HIGH_VALUE,
                            "m_stock = m_stock + " + detail.getConfirmQuantity())
                    .setSql(detail.getAssignRule() == AssignRuleEnum.LOW_VALUE,
                            "m_stock = m_stock + " + detail.getQuantity())
            ));
        }
    }


    @PostMapping("audit-apply")
    @Operation(summary = "提交盘点申请单")
    @Transactional(rollbackFor = Exception.class)
    public void auditApply(@Validated @Parameter(description = "申请单") @RequestBody AuditApplyInfo applyInfo) {
        ApplyForm applyForm = applyInfo.to();
        applyForm.setType(ApplyTypeEnum.AUDIT);
        applyFormService.save(applyForm);

        List<Material> materials = materialService.listByIds(applyInfo.getIds());

        materials.forEach(material -> applyDetailService.save(ApplyDetail.builder()
                .applyId(applyForm.getId())
                .materialId(material.getId())
                .quantity(material.getStock())
                .build()));
    }
    @PostMapping("scan-data/{applyId}")
    @Operation(summary = "提交盘点扫码数据")
    @Transactional(rollbackFor = Exception.class)
    public void saveScanData(@Parameter(description = "扫码数据列表") @RequestBody List<StocktakingScanDetail> scanData,
                             @Parameter(description = "申请单ID") @PathVariable(name = "applyId") Long applyId){
        //保存扫码数据
        stocktakingScanDetailService.saveBatch(scanData);
        //修改状态
        applyFormService.update(Wrappers.<ApplyForm>lambdaUpdate()
                        .set(ApplyForm::getReviewResult, ReviewResultEnum.WAIT_REVIEW)
                        .eq(ApplyForm::getId, applyId));
        new Thread(() -> {
            comparisonData(scanData, applyId);
        }).start();
    }

    // 比较数据 生成差异数据
    public void comparisonData(List<StocktakingScanDetail> scanData, Long applyId){
        //生成map比较
        Map<Long, List<String>> scanMap = scanData.stream().collect(Collectors.groupingBy(StocktakingScanDetail::getMaterialId,
                Collectors.mapping(StocktakingScanDetail::getBarcode, Collectors.toList())));
        Map<Long, List<MaterialStockDetail>> stockMap = getStockMap(applyId);
        // 比较两个map,以库存为准,生成差异数据
        List<StocktakingScanExceptionalData> exceptionalDataList = new ArrayList<>();
        scanMap.forEach((k, v) -> {
            List<MaterialStockDetail> stockList = stockMap.get(k);
            List<String> scanList = scanMap.get(k);
            // 比较两个list,找出库存存在但扫码不存在的条形码 或者 库存不存在但扫码存在的条形码
            stockList.forEach(detail -> {
                if (scanList.contains(detail.getBarcode()) && detail.getStatus() != StockStatusEnum.IN) {
                    // 扫码和库存都存在，但是库存的状态不对
                    exceptionalDataList.add(StocktakingScanExceptionalData.builder()
                            .barcode(detail.getBarcode())
                            .materialId(detail.getMaterialId())
                            .exceptionReason(StocktakingExceptonalStatusEnum.SOCK_OUT_BUT_SCAN_EXIST)
                            .build());
                }
                if(!scanList.contains(detail.getBarcode()) && detail.getStatus() == StockStatusEnum.IN){
                    //库房存在，但扫码不存在
                    exceptionalDataList.add(StocktakingScanExceptionalData.builder()
                            .barcode(detail.getBarcode())
                            .materialId(detail.getMaterialId())
                            .exceptionReason(StocktakingExceptonalStatusEnum.SOCK_IN_BUT_SCAN_NOT_EXIST)
                            .build());
                }
            });
            // 找出扫码存在但库存不存在的条形码
            scanList.forEach(barcode -> {
                if (stockList.stream().noneMatch(detail -> detail.getBarcode().equals(barcode))) {
                    // 扫码存在，但库存不存在
                    exceptionalDataList.add(StocktakingScanExceptionalData.builder()
                            .barcode(barcode)
                            .materialId(k)
                            .exceptionReason(StocktakingExceptonalStatusEnum.SOCK_NOT_EXIST_BUT_SCAN_EXIST)
                            .build());
                }
            });
        });
        // 保存差异数据
        stocktakingScanExceptionalDataService.saveBatch(exceptionalDataList);
    }
    // 盘点物料库存实际数据
    private Map<Long, List<MaterialStockDetail>> getStockMap(Long applyId) {
        ApplyForm applyForm = applyFormService.getById(applyId);
        List<Long> materialIds = new ArrayList<>();
        if(applyForm.getAuditType() == AuditTypeEnum.PARTIAL){
            // 部分盘点
            List<ApplyDetail> applyDetails = applyDetailService.list(Wrappers.<ApplyDetail>lambdaQuery().eq(ApplyDetail::getApplyId, applyId));
            materialIds = applyDetails.stream().map(ApplyDetail::getMaterialId).toList();
        }
        //查询所有盘点物料数量
        List<MaterialStockDetail> stockDetails = materialStockDetailService.list(Wrappers.<MaterialStockDetail>lambdaQuery()
                .in(!materialIds.isEmpty(), MaterialStockDetail::getMaterialId, materialIds));

        // 返回map结构 key 为物料id，value 该物料的所有条形码结合
        return stockDetails.stream().collect(Collectors.groupingBy(MaterialStockDetail::getMaterialId));
    }
}
